/*
 *  AVRVM, a virtual machine which emulates an Atmel AVR CPU.
 *  Copyright (C) 2010  Joakim Soderlund
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>

#include "instructions.h"

#ifdef DEBUG
#define dprintf(format, ...) printf("%-11s - ", __FUNCTION__);printf(format, __VA_ARGS__)
#else
#define dprintf(format, ...)
#endif

AVRVMInstruction avrvm_instructions[] = {
	{ AVRVM_LDI_FILTER,   AVRVM_LDI_MASK,   avrvm_ldi   },
	{ AVRVM_OUT_FILTER,   AVRVM_OUT_MASK,   avrvm_out   },
	{ AVRVM_IN_FILTER,    AVRVM_IN_MASK,    avrvm_in    },
	{ AVRVM_AND_FILTER,   AVRVM_AND_MASK,   avrvm_and   },
	{ AVRVM_ANDI_FILTER,  AVRVM_ANDI_MASK,  avrvm_andi  },
	{ AVRVM_OR_FILTER,    AVRVM_OR_MASK,    avrvm_or    },
	{ AVRVM_ORI_FILTER,   AVRVM_ORI_MASK,   avrvm_ori   },
	{ AVRVM_RJMP_FILTER,  AVRVM_RJMP_MASK,  avrvm_rjmp  },
	{ AVRVM_SEZ_FILTER,   AVRVM_SEZ_MASK,   avrvm_sez   },
	{ AVRVM_SEV_FILTER,   AVRVM_SEV_MASK,   avrvm_sev   },
	{ AVRVM_PUSH_FILTER,  AVRVM_PUSH_MASK,  avrvm_push  },
	{ AVRVM_POP_FILTER,   AVRVM_POP_MASK,   avrvm_pop   },
	{ AVRVM_RCALL_FILTER, AVRVM_RCALL_MASK, avrvm_rcall },
	{ AVRVM_RET_FILTER,   AVRVM_RET_MASK,   avrvm_ret   },
	{ AVRVM_NOP_FILTER,   AVRVM_NOP_MASK,   avrvm_nop   },
	{ AVRVM_BREQ_FILTER,  AVRVM_BREQ_MASK,  avrvm_breq  },
	{ AVRVM_BRNE_FILTER,  AVRVM_BRNE_MASK,  avrvm_brne  },
 //	{ AVRVM_CP_FILTER,    AVRVM_CP_MASK,    avrvm_cp    },   Not implemented yet.
	{ 0,                  0,                0           } // Null terminator, sort of.
};


/* LDI - Load imidiate -  */
int avrvm_ldi(ATMega16 *cpu) {
	unsigned char gen_register = 16 + ((cpu->ir & 0x00F0) >> 4);
	unsigned char value = ((cpu->ir & 0x0F00) >> 4) + ((cpu->ir & 0x000F));
	
	cpu->gen[gen_register] = value;
	
	dprintf("%04X - LDI   R%2d, %3d\n", cpu->ir, gen_register, value);
	return 0;
}


/* AND - Bitwise and with immidiate. */
int avrvm_and(ATMega16 *cpu) {
	unsigned char register_target = ((cpu->ir & 0x01F0) >> 4);
	unsigned char register_source = ((cpu->ir & 0x0200) >> 5) + (cpu->ir & 0x000F);
	
	unsigned char result = cpu->gen[register_target] & cpu->gen[register_source];
	
	cpu->gen[register_target] = result;
	
	// Clear S, V, N and Z.
	*cpu->sreg &= ~((1 << SREG_S) | (1 << SREG_V) | (1 << SREG_N) | (1 << SREG_Z));
	
	// Set N to MSB.
	*cpu->sreg |= ((result >> 7) << SREG_N);
	
	// Set S to N XOR V.
	*cpu->sreg |= ((((*cpu->sreg >> SREG_N) & 1) ^ ((*cpu->sreg >> SREG_V) & 1)) << SREG_S);
	
	// Set Z if result == 0;
	if (result == 0)
		*cpu->sreg |= (1 << SREG_Z);

	dprintf("%04X - AND   R%2d, R%2d\n", cpu->ir, register_target, register_source);
	return 0;
}


/* OR - Bitwise or. */
int avrvm_or(ATMega16 *cpu) {
	unsigned char register_target = ((cpu->ir & 0x01F0) >> 4);
	unsigned char register_source = ((cpu->ir & 0x0200) >> 5) + (cpu->ir & 0x000F);
	
	unsigned char result = cpu->gen[register_target] | cpu->gen[register_source];
	
	cpu->gen[register_target] = result;
	
	// Clear S, V, N and Z.
	*cpu->sreg &= ~((1 << SREG_S) | (1 << SREG_V) | (1 << SREG_N) | (1 << SREG_Z));
	
	// Set N to MSB.
	*cpu->sreg |= ((result >> 7) << SREG_N);
	
	// Set S to N XOR V.
	*cpu->sreg |= ((((*cpu->sreg >> SREG_N) & 1) ^ ((*cpu->sreg >> SREG_V) & 1)) << SREG_S);
	
	// Set Z if result == 0;
	if (result == 0)
		*cpu->sreg |= (1 << SREG_Z);

	dprintf("%04X - AND   R%2d, R%2d\n", cpu->ir, register_target, register_source);
	return 0;
}


/* ANDI - Bitwise and with immidiate. */
int avrvm_ori(ATMega16 *cpu) {
	unsigned char gen_register = 16 + ((cpu->ir & 0x00F0) >> 4);
	unsigned char value = ((cpu->ir & 0x0F00) >> 4) + (cpu->ir & 0x000F);
	
	unsigned char result = cpu->gen[gen_register] & value;
	
	cpu->gen[gen_register] = result;
	
	// Clear S, V, N and Z.
	*cpu->sreg &= ~((1 << SREG_S) | (1 << SREG_V) | (1 << SREG_N) | (1 << SREG_Z));
	
	// Set N to MSB.
	*cpu->sreg |= ((result >> 7) << SREG_N);
	
	// Set S to N XOR V.
	*cpu->sreg |= ((((*cpu->sreg >> SREG_N) & 1) ^ ((*cpu->sreg >> SREG_V) & 1)) << SREG_S);
	
	// Set Z if result == 0;
	if (result == 0)
		*cpu->sreg |= (1 << SREG_Z);

	dprintf("%04X - ANDI  R%2d, %3d\n", cpu->ir, gen_register, value);
	return 0;
}


/* ORI - Bitwise and with immidiate. */
int avrvm_andi(ATMega16 *cpu) {
	unsigned char gen_register = 16 + ((cpu->ir & 0x00F0) >> 4);
	unsigned char value = ((cpu->ir & 0x0F00) >> 4) + (cpu->ir & 0x000F);
	
	unsigned char result = cpu->gen[gen_register] | value;
	
	cpu->gen[gen_register] = result;
	
	// Clear S, V, N and Z.
	*cpu->sreg &= ~((1 << SREG_S) | (1 << SREG_V) | (1 << SREG_N) | (1 << SREG_Z));
	
	// Set N to MSB.
	*cpu->sreg |= ((result >> 7) << SREG_N);
	
	// Set S to N XOR V.
	*cpu->sreg |= ((((*cpu->sreg >> SREG_N) & 1) ^ ((*cpu->sreg >> SREG_V) & 1)) << SREG_S);
	
	// Set Z if result == 0;
	if (result == 0)
		*cpu->sreg |= (1 << SREG_Z);

	dprintf("%04X - ANDI  R%2d, %3d\n", cpu->ir, gen_register, value);
	return 0;
}


/* RJMP - Relative jump. */
int avrvm_rjmp(ATMega16 *cpu) {
	short steps = (((cpu->ir & 0x0800) >> 11) * -2048) + (cpu->ir & 0x07FF);
	cpu->pc += steps;
	
	dprintf("%04X - RJMP  %3d\n", cpu->ir, steps);
	return 0;
}


/* SEZ - Set the Z flag in SREG. */
int avrvm_sez(ATMega16 *cpu) {
	*cpu->sreg |= (1 << SREG_Z);
	
	dprintf("%04X - SEZ\n", cpu->ir);
	return 0;
}


/* SEV - Set the V flag in SREG. */
int avrvm_sev(ATMega16 *cpu) {
	*cpu->sreg |= (1 << SREG_V);
	
	dprintf("%04X - SEV\n", cpu->ir);
	return 0;
}


/* OUT - Write general register contents to IO register. */
int avrvm_out(ATMega16 *cpu) {
	unsigned char gen_register = ((cpu->ir & 0x01F0) >> 4);
	unsigned char io_register = ((cpu->ir & 0x0600) >> 5) + (cpu->ir & 0x000F);
	
	cpu->io[io_register] = cpu->gen[gen_register];
	
	dprintf("%04X - OUT   %3d, R%2d\n", cpu->ir, io_register, gen_register);
	return 0;
}


/* IN - Write IO register contents to general register. */
int avrvm_in(ATMega16 *cpu) {
	unsigned char gen_register = ((cpu->ir & 0x01F0) >> 4);
	unsigned char io_register = ((cpu->ir & 0x0600) >> 5) + (cpu->ir & 0x000F);
	
	cpu->gen[gen_register] = cpu->io[io_register];
	
	dprintf("%04X - IN    R%2d, %3d\n", cpu->ir, gen_register, io_register);
	return 0;
}


/* PUSH - Push the contents of a general register to the stack. */
int avrvm_push(ATMega16 *cpu) {
	unsigned char gen_register = ((cpu->ir & 0x01F0) >> 4);
	
	cpu->sram[ntohs(*cpu->sp)] = cpu->gen[gen_register];
	*cpu->sp = htons(ntohs(*cpu->sp) - 1);
	
	dprintf("%04X - PUSH  R%2d\n", cpu->ir, gen_register);
	return 0;
}


/* POP - Pop the contents of a the stack to a general register. */
int avrvm_pop(ATMega16 *cpu) {
	unsigned char gen_register = ((cpu->ir & 0x01F0) >> 4);
	
	*cpu->sp = htons(ntohs(*cpu->sp) + 1);
	cpu->gen[gen_register] = cpu->sram[ntohs(*cpu->sp)];
	
	dprintf("%04X - POP   R%2d\n", cpu->ir, gen_register);
	return 0;
}


/* RCALL - Calls a subroutine. */
int avrvm_rcall(ATMega16 *cpu) {
	short steps = (((cpu->ir & 0x0800) >> 11) * -2048) + (cpu->ir & 0x07FF);
	
	cpu->sram[ntohs(*cpu->sp)] = cpu->pc;
	*cpu->sp = htons(ntohs(*cpu->sp) - 1);
	
	cpu->pc += steps;

	dprintf("%04X - RCALL %3d\n", cpu->ir, steps);
	return 0;
}


/* RET - Returns from a subroutine. */
int avrvm_ret(ATMega16 *cpu) {
	*cpu->sp = htons(ntohs(*cpu->sp) + 1);
	cpu->pc = cpu->sram[ntohs(*cpu->sp)];
	
	dprintf("%04X - RET\n", cpu->ir);
	return 0;
}


/* NOP - No operation, the instructions, they do nothing. */
int avrvm_nop(ATMega16 *cpu) {
	dprintf("%04X - NOP\n", cpu->ir);
	return 0;
}


/* BREQ - Branch if equal. */
int avrvm_breq(ATMega16 *cpu) {
	short steps = (((cpu->ir & 0x0200) >> 9) * -64) + ((cpu->ir & 0x01F8) >> 3);

	if (*cpu->sreg & (1 << SREG_Z)) {
		cpu->pc += steps;
	}
	
	dprintf("%04X - BREQ  %3d", cpu->ir, steps);
	return 0;
}


/* BRNE - Branch if not equal. */
int avrvm_brne(ATMega16 *cpu) {
	short steps = (((cpu->ir & 0x0200) >> 9) * -64) + ((cpu->ir & 0x01F8) >> 3);

	if (!(*cpu->sreg & (1 << SREG_Z))) {
		cpu->pc += steps;
	}
	
	dprintf("%04X - BREQ  %3d", cpu->ir, steps);
	return 0;
}


/* CP - Compare - Not complete yet. */
int avrvm_cp(ATMega16 *cpu) {
	unsigned char register_left = ((cpu->ir & 0x01F0) >> 4);
	unsigned char register_right = ((cpu->ir & 0x0200) >> 5) + (cpu->ir & 0x000F);
	
	unsigned char result = cpu->gen[register_left] - cpu->gen[register_right];
	
	// Clear H, S, V, N, Z and C.
	*cpu->sreg &= ~((1 << SREG_H) | (1 << SREG_S) | (1 << SREG_S) | (1 << SREG_V) | (1 << SREG_N) | (1 << SREG_Z) | (1 << SREG_C));
	
	// Set Z flag if reult == 0.
	if (result == 0) {
		*cpu->sreg |= (1 << SREG_Z);
	}
}

